/*
 * vRouter: router.h
 *
 * Purpose: Declare classes for Routers, DataLinks, and some
 *          utility objects.
 *
 * Changes:
 *  0.02 - Removed constants for array sizes, uses C++ vectors instead
 *  0.01 - Initial Release
 *
 */

#ifndef ROUTER__H
#define ROUTER__H

#include <string>
#include <iostream>

/* Declare these classes first so that the other includes have them */

class DataLink; // virtual datalinks
class Router;  // virtugal routers
class Qentry; // a structure for outbound packet queues

#include "network.h"
#include "ipaddress.h"
#include "packet.h"
#include <queue>
#include <list>

#include <vector>
 
#include <string.h>

/* Note: the constants that were here on 0.01 are gone.  Everything
   uses C++ vectors now so the arbitrary limits on network size, etc.
   are gone.  Yaay!
*/

/* A class for virtual datalink functionality */
class DataLink {
  friend Router;

public:
	DataLink(IPAddress network, int bandwidth=1000, int delay=100);
	~DataLink();

	void addRouter(Router *rtr, IPAddress a);
	void removeRouter(Router *rtr);
	IPAddress net();
	void listHosts();
	void deliverPacket(Packet *p);  // shouldn't be used anymore
	void deliverPacket(Packet *p, IPAddress nextHop);

	void runTimeSlice(int timeWidth); // delivers all packets in Q
	int bandwidth() { return bandwidth_;};
	int delay() { return delay_;};
	
 private:
	IPAddress net_; // IP network address of this datalink

	vector <Router*> routers_; // all the routers attached
	vector <IPAddress*> ipaddresses_; // the routers IP addresses

	int bandwidth_; // bandwidth of the link in bps
	int delay_;     // delay of the link in ms
	
	vector< Qentry> Q;
};


/* A class which implements router-like functionality */
class Router {
  friend DataLink;
public: 
	Router(string name);
	~Router();

	void addInterface(DataLink* network, IPAddress address);
	void printInterfaces();
	string name() { return name_; } ;
	
	int destInterface(IPAddress a);
	void addRoute(IPAddress dst, IPAddress nextHop, int metric);
	void printRoutes();
	int localInterface(IPAddress a);
	
	void deliverPacket(Packet *p);
	void runTimeSlice(int time);
	void rippleQueue();

	void spfRoute(Network *);


private:
	void handleLocalPacket(Packet *p);
	int  servicePacket(Packet *p);
	IPAddress nextHop(IPAddress a);
	void queuePacket(Packet *p);
	void delayQueue(list<Qentry>, int timeWidth);

	bool forMe(Packet *p);

	vector< DataLink *> dataLinks_;
	vector <IPAddress*> IPAddresses_;

	string name_;

	vector<Packet*> recvQueue;

	vector< list <Qentry> > sendQ;
	queue<Packet*> localQ;

	/* Each entry in the routing table consists of 2 IPAddress 
	   objects, the first is the destination address, and the second
	   is the appropriate next hop address */
	vector < vector<IPAddress> > routeTable_;

	vector< int > linkQuotas;
};


/* Used to store packets in outbound queues. */
class Qentry
{

 public:
  Qentry(Packet *pack, IPAddress hop) : p(pack), nxHop(hop) { };
  Qentry() { } ;

  Packet *p;
  IPAddress nxHop;
};

#endif

